####################################################################################
#                                                                                  #
# Copyright (c) 2006 Dr. Conan C. Albrecht <conan_albrechtATbyuDOTedu>             #
#                                                                                  #
# This file is part of Picalo.                                                     #
#                                                                                  #
# Picalo is free software; you can redistribute it and/or modify                   #
# it under the terms of the GNU General Public License as published by             #
# the Free Software Foundation; either version 2 of the License, or                # 
# (at your option) any later version.                                              #
#                                                                                  #
# Picalo is distributed in the hope that it will be useful,                        #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                   #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                    #
# GNU General Public License for more details.                                     #
#                                                                                  #
# You should have received a copy of the GNU General Public License                #
# along with Foobar; if not, write to the Free Software                            #
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA        #
#                                                                                  #
####################################################################################
#                                                                                  #
#  This file is globally imported into Picalo.  See picalo/__init__.py.            # 
#  When you write "from picalo import *", these functions all get imported.        #
#                                                                                  #
####################################################################################


import os, sys, time, types, codecs, gzip
import re, inspect, linecache, csv, xml.dom.minidom
from Table import Table
from picalo import check_valid_table, show_progress, clear_progress, ensure_valid_variables, make_unique_colnames
import ZODB.DB
import persistent
from persistent.list import PersistentList


# global functions (defined in this file)
__all__ = [
  'TableArray',
]
  

##########################################################
###   TableArray type
###   A simple extension to list to allow lists of tables
###   to be identified without having to look through all
###   the elements to ensure they are Tables.
###
###   All of the elements of a TableArray must be 
###   compatible with the same columns, etc.  

class TableArray(persistent.Persistent):
  def __init__(self, *args):
    '''A list of picalo tables.  Some functions in Picalo return a set
       of tables.  Sets of tables are returned as TableArray objects.  These
       objects have the exact same methods and behavior as typical Python
       lists.
       
       For example, the Grouping.stratify_by_value() method stratifies
       a table into a number of smaller tables.  These tables are
       returned in TableArray objects.  You can access individual tables
       in a TableArray using the [n] notation.
       
       Users do not normally create TableArrays.  They are created automatically
       by Picalo functions.
       
       TableArrays must hold the exact same type of table.  For example, two
       tables in a given TableArray may not have different column names.
       If you need to hold a list of noncompatible tables (i.e. tables with
       different columns or types), use a TableList object.
       
       Example:
        >>> data = Table([
        ...   ('id', int),
        ...   ('name', unicode),
        ... ],[
        ...   [ 1, 'Benny' ],
        ...   [ 2, 'Vijay' ],
        ... ])
        >>> tables = Grouping.stratify_by_value(data, 'id')
        >>> tables[0].view()
        +----+-------+
        | id |  name |
        +----+-------+
        |  1 | Benny |
        +----+-------+
        >>> tables[1].view()
        +----+-------+
        | id |  name |
        +----+-------+
        |  2 | Vijay |
        +----+-------+
        >>> for table in tables:
        ...   print len(table)        
        ...
        1
        1
    '''
    self.changed = False
    if len(args) == 0:
      self._tables = PersistentList() 
    else:
      self._tables = PersistentList(args[0]) 
    for value in self:
      self._check_valid_table(value)
      
  
  def view(self):
    '''Opens the table list for viewing in the Picalo user interface.
       The resulting view allows you to page through the tables in the
       list.  See the first example.
       
       You can view individual tables in the list by using the [n]
       notation.  See the second example for this notation.
       
       Example:
        >>> data = Table([
        ...   ('id', int),
        ...   ('name', unicode),
        ... ],[
        ...   [ 1, 'Benny' ],
        ...   [ 2, 'Vijay' ],
        ... ])
        >>> tables = Grouping.stratify_by_value(data, 'id')
        >>> tables.view()
         
       Example:
        >>> data = Table([
        ...   ('id', int),
        ...   ('name', unicode),
        ... ],[
        ...   [ 1, 'Benny' ],
        ...   [ 2, 'Vijay' ],
        ... ])
        >>> tables = Grouping.stratify_by_value(data, 'id')
        >>> tables[0].view()
         
    '''
    from picalo import mainframe
    if mainframe != None:
      # try to figure out the variable name of this table by inspecting the call stack
      f = inspect.currentframe().f_back
      locals = f.f_locals
      for name in f.f_code.co_names:
        if locals.has_key(name) and locals[name] == self:
          mainframe.openTable(name)
          return
    # if we get here, either we're in console mode or we couldn't find the variable name
    print self
    

  def _check_valid_table(self, table):
    '''Ensures that a given table is valid'''
    try:
      assert isinstance(table, (Table, TableArray)), 'This type of list can only hold Picalo Tables or TableArrays.  You sent %s.' % (table.__class__, )
      if len(self) > 0:
        assert table.get_columns() == self[0].get_columns(), 'Incompatible table.  All tables in a TableArray must have the same columns, types, etc.'
    except AssertionError, e:
      raise TypeError, str(e)
    
  
  def __repr__(self):
    '''For debugging.
       @return:   The number of tables in this TableArray.
       @rtype:    str
    '''
    return '<TableArray of %s: %s>' % (len(self), ', '.join([ repr(t) for t in self._tables]))
  
  def __str__(self):
    '''For debugging.
       @return:   The number of tables in this TableArray.
       @rtype:    str
    '''
    return '<TableArray of %s: %s>' % (len(self), ', '.join([ str(t) for t in self._tables]))
  
  
  ##### CONTAINER METHODS ####
  
  def __iter__(self):
    '''Returns an iterator to this TableArray'''
    return TableArrayIterator(self)
    
    
  def __contains__(self, table):
    '''Returns whether the given table is in this TableArray'''
    if isinstance(table, (Table, TableArray)):
      for item in self:
        if item.get_columns() == table.get_columns() and item == table: # item == table checks the values of the tables
          return true
    return False
    
    
  def __eq__(self, other):
    '''Returns whether this TableArray is equal to a list/tuple/TableArray of tables.'''
    if other != None and len(self) == len(other):
      for i in range(len(self)):
        if self[i] != other[i]:
          return False
      return True
    return False
    

  def __ne__(self, other):
    '''Returns whether this TableArray is not equal to a list/tuple/TableArray of tables.'''
    return not self.__eq__(other)  
  
  
  def __len__(self):
    return len(self._tables)
    
  
  def __getitem__(self, index):
    return self._tables[index]


  def __setitem__(self, index, table):
    self._check_valid_table(table)
    self.changed = True
    self._tables[index] = table

  
  def __setslice__(self, i, j, sequence):
    for value in sequence:
      self._check_valid_table(value)
    self.changed = True
    self._table.__setslice__(i, j, sequence)


  def __getslice__(self, i, j):  
    return self._table.__getslice__(i, j)
    
    
  def append(self, table):
    self._check_valid_table(table)
    self.changed = True
    self._tables.append(table)

    

  ###################################################################
  ###   Table functions that are applied to each table in the list

  def _add_listener(self, listener):
    '''Adds a listener to the table.  Will be notified when data changes occur.
       The listener should be a callable/function of form callback(table).  For
       efficiency reasons and because we don't need it right now,
       the col and row is not reported.'''
    for table in self:
      table._add_listener(listener)
    
    
  def _remove_listener(self, listener):
    '''Removes a listener from the table.'''
    for table in self:
      table._remove_listener(listener)


  def _notify_listeners(self, level=1):
    '''Notifies listeners that we've had a change'''
    for table in self:
      table._notify_listeners(level)


  def _invalidate_indexes(self):
    '''Invalidates the indices already calculated on this table.  This occurs anytime
       data are modified in this table'''
    for table in self:
      table._invalidate_indexes()


  def is_changed(self):
    '''Returns whether the table has been changed since loading'''
    if self.changed:
      return True
    for table in self:
      if table.is_changed():
        return True
    return False
    
    
  def set_changed(self, changed):
    '''Sets whether the class has been changed since loading.  This is not normally
       called by users.'''
    self.changed = True


  def set_readonly(self, readonly_flag=False):
    '''Sets the read only status of this table.  Tables that are read only cannot
       be modified.  Normally, tables are initially not read only (i.e. can be modified).
       The only exception is tables loaded from databases, which are read only.
       
       @param readonly_flag: True or False, depending upon whether the table should be read only or not.
       @type  readonly_flag: bool
    '''
    for table in self:
      table.set_readonly(readonly_flag)
  

  def is_readonly(self):
    '''Returns whether this table is read only.
       @return:  Whether this table is read only.
       @rtype:   bool
    '''
    if len(self) > 0:
      return self[0].is_readonly()
    return False


  def filter(self, expression=None):
    '''Applies the given filter to all tables in the list.  
       See the definition of this method in Table for more 
       information.'''
    for table in self:
      table.filter(expression)
      
      
  def clear_filter(self):
    '''Clears any active filter to all tables in the list.  
       See the definition of this method in Table for more 
       information.'''
    for table in self:
      table.clear_filter()
       

  def is_filtered(self):
    '''Returns whether the first table in this list is filtered.'''
    if len(self) > 0:
      return self[0].is_filtered()
    return False


  def get_filter_expression(self):
    '''Returns the filter expression as a PicaloExpression object, or None if no filter is applied.'''
    if len(self) > 0:
      return self[0].get_filter_expression()
    return None


  def column(self, col):
    '''Returns the given column from the first table in this list.  Since all
       tables have the same column definitions, it is comparable across all
       tables in the list.'''
    if len(self) > 0:
      return self[0].column(col)
    return None
    

  def get_columns(self):
    '''Returns the columns from the first table in this list.  Since all
       tables have the same column definitions, it is comparable across all
       tables in the list.'''
    if len(self) > 0:
      return self[0].get_columns()
    return None
    
    
  def get_column_names(self):
    '''Returns the column names from the first table in this list.  Since all
       tables have the same column definitions, it is comparable across all
       tables in the list.'''
    if len(self) > 0:
      return self[0].get_column_names()
    return []
    
    
  def column_count(self):
    '''Returns the column count from the first table in this list.  Since all
       tables have the same column definitions, it is comparable across all
       tables in the list.'''
    if len(self) > 0:
      return self[0].column_count()
    return 0


  def set_name(self, column, name):
    '''Sets the column name for all tables in the list.  
       See the definition of this method in Table for more 
       information.'''
    for table in self:
      table.set_name(column, name)


  def set_type(self, column, column_type=None, format=None, expression=None):
    '''Sets the column type for all tables in the list.  
       See the definition of this method in Table for more 
       information.'''
    for table in self:
      table.set_type(column, column_type, format, expression)


  def set_format(self, column, format=None):
    '''Sets the format for a column in all tables in the list.
       See the definition of this method in Table for more information.'''
    for table in self:
      table.set_format(column, format)       
    
    
  def guess_types(self, num_records=-1):
    '''Guesses the column types for all tables in the list.  
       See the definition of this method in Table for more 
       information.'''
    for table in self:
      table.guess_types(num_records)


  def append_column(self, name, column_type, values=None):
    '''Adds a new column to each table in the list.
       See the definition of this method in Table for more 
       information.'''
    assert isinstance(values, (types.NoneType, TableArray)), 'If the values parameter is provided, it must be a TableArray of values to add'
    for i, table in enumerate(self):
      if values == None or i >= len(values):
        table.append_column(name, column_type)
      else:
        table.append_column(name, column_type, values[i])
      
      
  def insert_column(self, index, name, column_type, values=None):
    '''Inserts a new column to each table in the list.
       See the definition of this method in Table for more 
       information.'''
    assert isinstance(values, (types.NoneType, TableArray)), 'If the values parameter is provided, it must be a TableArray of values to add'
    for i, table in enumerate(self):
      if values == None or i >= len(values):
        table.insert_column(name, column_type)
      else:
        table.insert_column(index, name, column_type, values[i])
      

  def append_calculated(self, name, expression):
    '''Appends a calculated column to each table in the list.
       See the definition of this method in Table for more 
       information.'''
    for table in self:
      table.append_calculated(name, expression)
  
  
  def insert_calculated(self, index, name, expression):
    '''Inserts a calculated column to each table in the list.
       See the definition of this method in Table for more 
       information.'''
    for table in self:
      table.insert_calculated(index, name, expression)
  
  
  def append_calculated_static(self, name, column_type, expression):
    '''Appends a calculated column to each table in the list.
       See the definition of this method in Table for more 
       information.'''
    for table in self:
      table.append_calculated_static(name, column_type, expression)
  

  def insert_calculated_static(self, index, name, column_type, expression):
    '''Inserts a calculated column to each table in the list.
       See the definition of this method in Table for more 
       information.'''
    for table in self:
      table.insert_calculated_static(index, name, column_type, expression)
  

  def move_column(self, column, new_index):
    '''Moves a column to another location for each table in the list.
       See the definition of this method in Table for more 
       information.'''
    for table in self:
      table.move_column(column, new_index)


  def delete_column(self, column):
    '''Deletes a column from each table in the list.
       See the definition of this method in Table for more 
       information.'''
    for table in self:
      table.delete_column(column)
  

  def structure(self):
    '''Returns the structure of the first table in this list.  Since all
       tables have the same column definitions, it is comparable across all
       tables in the list.'''
    if len(self) > 0:
      return self[0].structure()
    return None
    

  def combine(self):
    '''Combines this table array into a single table.  This is a kind of 'anti-stratification'.
       The new table has all the records of the tables in the TableArray.
    '''
    # create the new table
    results = Table(self.get_columns())
    # append the results from each table in the array
    try:
      for i, table in enumerate(self):
        show_progress('Combining into table...', float(i) / len(self))
        results.extend(table)
    finally:
      clear_progress()
    return results
    
          
  def save(self, filename, respect_filter=False):
    '''Saves this TableList in native Picalo format.  This is the preferred
       format to save TableLists in because all column types, formulas, and so
       forth are saved.
    
       @param filename: The filename to save to.  This can also be an open stream.
       @type  filename: str
       @param respect_filter Whether to save the entire file or only those rows available through any current filters.
       @type  respect_filter bool
    '''
    # I can't import save until now because TableArray class must be declared first
    from Table import save as table_save
    table_save(self, filename, respect_filter)


  def save_delimited(self, filename, delimiter=',', qualifier='"', line_ending='\n', none='', encoding='utf-8', respect_filter=False):
    '''Saves this TableList in delimited format.  Since the table list probably has multiple tables
       in it, one delimited file per table is created by prepending 1, 2, 3 to the filename.

       @param filename:     A file name or a file pointer to an open file.  Using the file name string directly is suggested since it ensures the file is opened correctly for reading in CSV.
       @type  filename:     str
       @param delimiter:    An optional field delimiter character
       @type  delimiter:    str
       @param qualifier:    An optional qualifier to use when delimiters exist in field values
       @type  qualifier:    str
       @param line_ending:  An optional line ending to separate rows with
       @type  line_ending:  str
       @param none:         An optional parameter specifying what to write for cells that have the None value.
       @type  none:         str
       @param encoding:     The unicode encoding to use for international or special characters.  For example, Microsoft applications like to use special characters for double quotes rather than the standard characters.  Unicode (the default) handles these nicely.
       @type  encoding:     str
       @param respect_filter Whether to save the entire file or only those rows available through any current filter.
       @type  respect_filter bool
    '''    
    for i, table in enumerate(self):
      table.save_delimited('%05i-' % i + filename, delimiter, qualifier, line_ending, none, encoding, respect_filter)
      

  def save_csv(self, filename, line_ending='\n', none='', encoding='utf-8', respect_filter=False):
    '''Saves this TableList in CSV format.  Since the table list probably has multiple tables
       in it, one CSV per table is created by prepending 1, 2, 3 to the filename.

       @param filename:     A file name or a file pointer to an open file.  Using the file name string directly is suggested since it ensures the file is opened correctly for reading in CSV.
       @type  filename:     str
       @param line_ending:  An optional line ending to separate rows with
       @type  line_ending:  str
       @param none:         An optional parameter specifying what to write for cells that have the None value.
       @type  none:         str
       @param encoding:     The unicode encoding to use for international or special characters.  For example, Microsoft applications like to use special characters for double quotes rather than the standard characters.  Unicode (the default) handles these nicely.
       @type  encoding:     str
       @param respect_filter Whether to save the entire file or only those rows available through any current filter.
       @type  respect_filter bool
    '''
    for i, table in enumerate(self):
      table.save_csv('%05i-' % i + filename, line_ending, none, encoding, respect_filter)
    

  def save_tsv(self, filename, line_ending='\n', none='', encoding='utf-8', respect_filter=False):
    '''Saves this TableList in tsv format.  Since the table list probably has multiple tables
       in it, one tsv file per table is created by prepending 1, 2, 3 to the filename.

       @param filename:     A file name or a file pointer to an open file.  Using the file name string directly is suggested since it ensures the file is opened correctly for reading in CSV.
       @type  filename:     str
       @param line_ending:  An optional line ending to separate rows with
       @type  line_ending:  str
       @param none:         An optional parameter specifying what to write for cells that have the None value.
       @type  none:         str
       @param encoding:     The unicode encoding to use for international or special characters.  For example, Microsoft applications like to use special characters for double quotes rather than the standard characters.  Unicode (the default) handles these nicely.
       @type  encoding:     str
    '''
    for i, table in enumerate(self):
      table.save_tsv('%05i-' % i + filename, line_ending, none, encoding, respect_filter)
    
    
  def save_fixed(self, filename, line_ending='\n', none='', respect_filter=False):
    '''Saves this TableList in fixed format.  Since the table list probably has multiple tables
       in it, one fixed file per table is created by prepending 1, 2, 3 to the filename.

       @param filename:     A file name or a file pointer to an open file.  Using the file name string directly is suggested since it ensures the file is opened correctly for reading in CSV.
       @type  filename:     str
       @param line_ending:  An optional line ending to separate rows with
       @type  line_ending:  str
       @param none:         An optional parameter specifying what to write for cells that have the None value.
       @type  none:         str
       @param respect_filter Whether to save the entire file or only those rows available through any current filter.
       @type  respect_filter bool
    '''
    for i, table in enumerate(self):
      table.save_fixed('%05i-' % i + filename, line_ending, none, respect_filter)
    
    
  def save_xml(self, filename, line_ending='\n', indent='\t', compact=False, none='', respect_filter=False):
    '''Saves this TableList in xml format.  Since the table list probably has multiple tables
       in it, one xml file per table is created by prepending 1, 2, 3 to the filename.

       @param filename:     A file name or a file pointer to an open file.  Using the file name string directly is suggested since it ensures the file is opened correctly for reading in CSV.
       @type  filename:     str
       @param line_ending:  An optional line ending to separate rows with (when compact is False)
       @type  line_ending:  str
       @param indent:       The character(s) to use for indenting (when compact is False)
       @type  indent:       str
       @param compact:      Whether to compact the XML or make it "pretty" with whitespace
       @type  compact:      bool
       @param none:         An optional parameter specifying what to write for cells that have the None value.
       @type  none:         str
       @param respect_filter Whether to save the entire file or only those rows available through any current filter.
       @type  respect_filter bool
    '''
    for i, table in enumerate(self):
      table.save_xml('%05i-' % i + filename, line_ending, indent, compact, none, respect_filter)
    
    
    
###############################
###   Iterator for a TableArray
   
def TableArrayIterator(ta):     
  '''Returns a generator object to iterate over the tables in a TableArray'''
  index = 0
  numcols = len(ta)
  while index < numcols:
    yield ta[index]
    index += 1  


    



#!/usr/bin/python

####################################################################################
#                                                                                  #
# Copyright (c) 2003 Dr. Conan C. Albrecht <conan_albrechtATbyuDOTedu>             #
#                                                                                  #
# This file is part of Picalo.                                                     #
#                                                                                  #
# Picalo is free software; you can redistribute it and/or modify                   #
# it under the terms of the GNU General Public License as published by             #
# the Free Software Foundation; either version 2 of the License, or                # 
# (at your option) any later version.                                              #
#                                                                                  #
# Picalo is distributed in the hope that it will be useful,                        #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                   #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOS E.  See the                    #
# GNU General Public License for more details.                                     #
#                                                                                  #
# You should have received a copy of the GNU General Public License                #
# along with Foobar; if not, write to the Free Software                            #
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA        #
#                                                                                  #
####################################################################################
#
# Version 1.0
# Oct 11, 2003: Now uses expressions everywhere (to match Picalo's interface elsewhere)
#               Much better documentation
# Mar 09, 2003: pivot() was compiling results backwards.  Fixed this.
# Dec 08, 2003: Version 1.0.  Started this log.
#               Updated to work with Tables
#
#
####################################################################################

from picalo import Table, TableArray, Simple, show_progress, clear_progress, check_valid_table
from picalo import run_tablearray, ensure_valid_variables, ensure_unique_list_value, make_unique_colnames, make_valid_variable
from Expression import PicaloExpression
import types


__doc__ = '''
The Crosstable module creates crosstables of data.  It is similiar to Excel's
powerful PivotTable function.  Crosstabling takes data that is in database format
and converts it to spreadsheet format, usually with summarization functions like
sum, average, etc.

The primary function of this module is pivot().  The other functions are more
advanced functions that show different levels of detail from the crosstabling
process.

Pivoting extremely large tables can take some time, depending upon your processor
speed and amount of memory.  Assuming you are running Picalo in GUI mode,
the functions will show a progress bar.
'''



# a listing of the public functions in this module (used by Manual.py)
__functions__ = (
  'pivot',
  'pivot_table',
  'pivot_map',
  'pivot_map_detail',
)




def pivot(table, col_fields, row_fields, expressions, delimiter='_'):
  '''Crosstables a Picalo table into a new table.  This routine is similar to
     Excel's PivotTable feature, but it is quite a bit more powerful (although
     not as easy to use).  
     
     The pivot function flattens the results into a two-dimensional table, with a 
     column for each expression of each data field.  Of all the pivot functions
     in the Crosstable module, this is the most like Excel's pivot table feature.
     
     There are no arbitrary limits on table size in Picalo, so you can pivot tables
     with many values (resulting in a significant number of rows and columns).
     The only limit on the number of resulting rows and columns is your memory.
     
     The first example shows a simple pivot on a single col_field, row_field,
     and single expression.  The second example shows a more complex pivot
     with multiple fields and expressions.  Note that the second example
     has some errors because the average function is being run on cells
     that have no values (divide by zero).
     
     Example 1:
        >>> # create a test table
        >>> data = Table([
        ...  ('Region', unicode),
        ...  ('Product', unicode),
        ...  ('Salesperson', unicode), 
        ...  ('Customer', unicode),
        ...  ('Amount', int),
        ... ],[
        ...  [ 'Rural','Computers',  'Mollie', 'Faiz',  500 ],
        ...  [ 'City',  'Monitors',  'Danny',  'Sheng', 700 ],
        ...  [ 'Rural', 'Mice',      'Mollie', 'Brian', 900 ],
        ...  [ 'City',  'Computers', 'Danny',  'Faiz',  300 ],
        ...  [ 'Rural', 'Monitors',  'Mollie', 'Sheng', 500 ],
        ...  [ 'Rural', 'Monitors',  'Mollie', 'Sheng', 500 ],
        ...  [ 'City',  'Mice',      'Danny',  'Brian', 100 ],
        ...  [ 'Rural', 'Monitors',  'Mollie', 'Faiz',  200 ],
        ...  [ 'City',  'Computers', 'Danny',  'Sheng', 400 ],
        ... ])
        >>> # perform the pivot
        >>> ret = Crosstable.pivot(data, 'Salesperson', 'Product', 'sum(group["Amount"])')
        >>> ret.view()     
        +-----------+-------+--------+--------+
        |   Pivot   | Danny | Mollie | Totals |
        +-----------+-------+--------+--------+
        | Computers |   700 |    500 |   1200 |
        | Mice      |   100 |    900 |   1000 |
        | Monitors  |   700 |   1200 |   1900 |
        | Totals    |  1500 |   2600 |   4100 |
        +-----------+-------+--------+--------+     

     Example 2:
        >>> # create a test table
        >>> data = Table([
        ...  ('Region', unicode),
        ...  ('Product', unicode),
        ...  ('Salesperson', unicode), 
        ...  ('Customer', unicode),
        ...  ('Amount', int),
        ... ],[
        ...  [ 'Rural','Computers',  'Mollie', 'Faiz',  500 ],
        ...  [ 'City',  'Monitors',  'Danny',  'Sheng', 700 ],
        ...  [ 'Rural', 'Mice',      'Mollie', 'Brian', 900 ],
        ...  [ 'City',  'Computers', 'Danny',  'Faiz',  300 ],
        ...  [ 'Rural', 'Monitors',  'Mollie', 'Sheng', 500 ],
        ...  [ 'Rural', 'Monitors',  'Mollie', 'Sheng', 500 ],
        ...  [ 'City',  'Mice',      'Danny',  'Brian', 100 ],
        ...  [ 'Rural', 'Monitors',  'Mollie', 'Faiz',  200 ],
        ...  [ 'City',  'Computers', 'Danny',  'Sheng', 400 ],
        ... ])
        >>> # perform the pivot
        >>> ret = Crosstable.pivot(data, ['Salesperson','Customer'], ['Product', 'Region'], ['sum(group["Amount"])','mean(group["Amount"])'])
        >>> ret.view()
        +-----------------+-------------------+-------------------+------------------+------------------+-------------------+-------------------+--------------------+--------------------+-------------------+-------------------+--------------------+--------------------+--------------+---------------+
        |      Pivot      | Danny_Brian_expr1 | Danny_Brian_expr2 | Danny_Faiz_expr1 | Danny_Faiz_expr2 | Danny_Sheng_expr1 | Danny_Sheng_expr2 | Mollie_Brian_expr1 | Mollie_Brian_expr2 | Mollie_Faiz_expr1 | Mollie_Faiz_expr2 | Mollie_Sheng_expr1 | Mollie_Sheng_expr2 | Totals_expr1 |  Totals_expr2 |
        +-----------------+-------------------+-------------------+------------------+------------------+-------------------+-------------------+--------------------+--------------------+-------------------+-------------------+--------------------+--------------------+--------------+---------------+
        | Computers_City  |                 0 |                 0 |              300 |              300 |               400 |               400 |                  0 |                  0 |                 0 |                 0 |                  0 |                  0 |          700 |         350.0 |
        | Computers_Rural |                 0 |                 0 |                0 |                0 |                 0 |                 0 |                  0 |                  0 |               500 |               500 |                  0 |                  0 |          500 |           500 |
        | Mice_City       |               100 |               100 |                0 |                0 |                 0 |                 0 |                  0 |                  0 |                 0 |                 0 |                  0 |                  0 |          100 |           100 |
        | Mice_Rural      |                 0 |                 0 |                0 |                0 |                 0 |                 0 |                900 |                900 |                 0 |                 0 |                  0 |                  0 |          900 |           900 |
        | Monitors_City   |                 0 |                 0 |                0 |                0 |               700 |               700 |                  0 |                  0 |                 0 |                 0 |                  0 |                  0 |          700 |           700 |
        | Monitors_Rural  |                 0 |                 0 |                0 |                0 |                 0 |                 0 |                  0 |                  0 |               200 |               200 |               1000 |              500.0 |         1200 |         400.0 |
        | Totals          |               100 |               100 |              300 |              300 |              1100 |             550.0 |                900 |                900 |               700 |             350.0 |               1000 |              500.0 |         4100 | 455.555555556 |
        +-----------------+-------------------+-------------------+------------------+------------------+-------------------+-------------------+--------------------+--------------------+-------------------+-------------------+--------------------+--------------------+--------------+---------------+

     @param table:         The Picalo table that will be crosstabled
     @type  table:         Table or TableArray
     @param col_fields:    A single field name or a list containing the field names of the fields used to group columns.
     @type  col_fields:    str/list
     @param row_fields:    A single field name or a list containing the field names of the fields used to group rows
     @type  row_fields:    str/list
     @param expressions:   A single expression or a list containing expressions used to do calculations on groupings; when the expression is evaluated, the keyword 'group' denotes the matching records for each cell
     @type  expressions:   str/list
     @param delimiter:     The character(s) to use to combine field values and expressions (for row and column headers)
     @type  delimiter:     str
     @return:              A new table containing the crosstabled results.  The last column and last row of the table contain the row and column totals
     @rtype:               Table
  '''
  if isinstance(table, TableArray):
    return run_tablearray('Running pivot...', pivot, table, col_fields, row_fields, expressions, delimiter=delimiter)
  # table and field type checking is performed in Crosstable.__init__
  assert isinstance(delimiter, types.StringType), 'Invalid delimiter: ' + str(delimiter)
  try:
    ct = Crosstabler(table, row_fields, col_fields, expressions)
    rowlabels = ct.get_row_labels()
    rowlabels.sort()
    collabels = ct.get_col_labels()
    collabels.sort()
    
    # header (the columns)
    show_progress('Running pivot...', 0.0/len(rowlabels))
    columns = [ make_valid_variable('_'.join(_ensure_tuple(row_fields))) ] # first column is for the row keys
    for collabel in collabels:
      colname = delimiter.join([ unicode(name) for name in _ensure_tuple(collabel)])
      for i in range(len(ct.expressions)):       
        columns.append(ensure_unique_list_value(columns, make_valid_variable(colname + (len(ct.expressions) != 1 and (delimiter + 'expr' + str(i+1)) or ''))))
    for i in range(len(ct.expressions)):
      columns.append(ensure_unique_list_value(columns, make_valid_variable('Totals' + (len(ct.expressions) != 1 and (delimiter + 'expr' + str(i+1)) or ''))))
    table = Table(zip(make_unique_colnames(columns), [ unicode for i in columns ]))
  
    # add the data by row 
    for k, rowlabel in enumerate(rowlabels):
      show_progress('Running pivot...', float(k)/len(rowlabels))
      columns = []
      columns.append(delimiter.join([ str(name) for name in _ensure_tuple(rowlabel)]))
      for collabel in collabels:
        values = ct.get_cell_calculations(rowlabel, collabel)
        for i in range(len(ct.expressions)):
          columns.append(values[i])
      # row totals
      for total in ct.get_row_calculations(rowlabel):
        columns.append(total)
      table.append(columns)
      
    # add rows for totals
    columns = [ 'Totals' ]
    for collabel in collabels:
      values = ct.get_col_calculations(collabel)
      for i in range(len(ct.expressions)):
        columns.append(values[i])
    # grand totals
    for calc in ct.get_table_calculations():
      columns.append(calc)
    table.append(columns)
    
    return table  
  finally:
    clear_progress()
    

  
  
   
def pivot_table(table, col_fields, row_fields, expressions):
  '''Crosstables a Picalo table into a new table.  This routine is similar to
     Excel's PivotTable feature, but it is quite a bit more powerful (although
     not as easy to use).  
     
     This version creates does not flatten results to a two-dimensional table
     like the pivot function.  It is a more advanced way of pivoting than
     the pivot function because it creates a list of expressions results for 
     each cell.  In other words, the results table is not normalized, but
     contains lists within each cell that contain the results.  When a single
     col_field, row_field, and expression is given, this function produces the
     exact same results as pivot().
     
     This way of pivoting is useful if you want a single column for each col
     field value and a single row for each row field value.  Even if you provide
     multiple expressions and/or multiple data fields, you'll still only get one
     col/row match for a given value set.  The multiple expressions on the multiple
     data fields will be contained in a list in the cell for each row/col match.
     
     There are no arbitrary limits on table size in Picalo, so you can pivot tables
     with many values (resulting in a significant number of rows and columns).
     The only limit on the number of resulting rows and columns is your memory.
     
     The first example shows a simple pivot on a single col_field, row_field,
     and single expression.  The second example shows a more complex pivot
     with multiple fields and expressions.  Note that the second example
     has some errors because the average function is being run on cells
     that have no values (divide by zero).
     
     Example 1:
        >>> # create a test table
        >>> data = Table([
        ...  ('Region', unicode),
        ...  ('Product', unicode),
        ...  ('Salesperson', unicode), 
        ...  ('Customer', unicode),
        ...  ('Amount', int),
        ... ],[
        ...  [ 'Rural','Computers',  'Mollie', 'Faiz',  500 ],
        ...  [ 'City',  'Monitors',  'Danny',  'Sheng', 700 ],
        ...  [ 'Rural', 'Mice',      'Mollie', 'Brian', 900 ],
        ...  [ 'City',  'Computers', 'Danny',  'Faiz',  300 ],
        ...  [ 'Rural', 'Monitors',  'Mollie', 'Sheng', 500 ],
        ...  [ 'Rural', 'Monitors',  'Mollie', 'Sheng', 500 ],
        ...  [ 'City',  'Mice',      'Danny',  'Brian', 100 ],
        ...  [ 'Rural', 'Monitors',  'Mollie', 'Faiz',  200 ],
        ...  [ 'City',  'Computers', 'Danny',  'Sheng', 400 ],
        ... ])
        >>> # perform the pivot
        >>> ret = Crosstable.pivot_table(data, 'Salesperson', 'Product', 'sum(group["Amount"])')
        >>> ret.view()
        +-----------+-------+--------+--------+
        |   Pivot   | Danny | Mollie | Totals |
        +-----------+-------+--------+--------+
        | Computers |   700 |    500 |   1200 |
        | Mice      |   100 |    900 |   1000 |
        | Monitors  |   700 |   1200 |   1900 |
        | Totals    |  1500 |   2600 |   4100 |
        +-----------+-------+--------+--------+        

     Example 2:
        >>> # create a test table
        >>> data = Table([
        ...  ('Region', unicode),
        ...  ('Product', unicode),
        ...  ('Salesperson', unicode), 
        ...  ('Customer', unicode),
        ...  ('Amount', int),
        ... ],[
        ...  [ 'Rural','Computers',  'Mollie', 'Faiz',  500 ],
        ...  [ 'City',  'Monitors',  'Danny',  'Sheng', 700 ],
        ...  [ 'Rural', 'Mice',      'Mollie', 'Brian', 900 ],
        ...  [ 'City',  'Computers', 'Danny',  'Faiz',  300 ],
        ...  [ 'Rural', 'Monitors',  'Mollie', 'Sheng', 500 ],
        ...  [ 'Rural', 'Monitors',  'Mollie', 'Sheng', 500 ],
        ...  [ 'City',  'Mice',      'Danny',  'Brian', 100 ],
        ...  [ 'Rural', 'Monitors',  'Mollie', 'Faiz',  200 ],
        ...  [ 'City',  'Computers', 'Danny',  'Sheng', 400 ],
        ... ])
        >>> # perform the pivot
        >>> ret = Crosstable.pivot_table(data, ['Salesperson','Customer'], ['Product', 'Region'], ['sum(group["Amount"])','mean(group["Amount"])'])
        >>> ret.view()
        +------------------------+--------------------+-------------------+--------------------+---------------------+--------------------+---------------------+----------------------------+
        |         Pivot          | ('Danny', 'Brian') | ('Danny', 'Faiz') | ('Danny', 'Sheng') | ('Mollie', 'Brian') | ('Mollie', 'Faiz') | ('Mollie', 'Sheng') |           Totals           |
        +------------------------+--------------------+-------------------+--------------------+---------------------+--------------------+---------------------+----------------------------+
        |  ('Computers', 'City') |             (0, 0) |        (300, 300) |         (400, 400) |              (0, 0) |             (0, 0) |              (0, 0) |               (700, 350.0) |
        | ('Computers', 'Rural') |             (0, 0) |            (0, 0) |             (0, 0) |              (0, 0) |         (500, 500) |              (0, 0) |                 (500, 500) |
        |       ('Mice', 'City') |         (100, 100) |            (0, 0) |             (0, 0) |              (0, 0) |             (0, 0) |              (0, 0) |                 (100, 100) |
        |      ('Mice', 'Rural') |             (0, 0) |            (0, 0) |             (0, 0) |          (900, 900) |             (0, 0) |              (0, 0) |                 (900, 900) |
        |   ('Monitors', 'City') |             (0, 0) |            (0, 0) |         (700, 700) |              (0, 0) |             (0, 0) |              (0, 0) |                 (700, 700) |
        |  ('Monitors', 'Rural') |             (0, 0) |            (0, 0) |             (0, 0) |              (0, 0) |         (200, 200) |       (1000, 500.0) |              (1200, 400.0) |
        | Totals                 |         (100, 100) |        (300, 300) |      (1100, 550.0) |          (900, 900) |       (700, 350.0) |       (1000, 500.0) | (4100, 455.55555555555554) |
        +------------------------+--------------------+-------------------+--------------------+---------------------+--------------------+---------------------+----------------------------+     
     
     @param table:         The Picalo table that will be crosstabled
     @type  table:         Table or TableArray
     @param col_fields:    A single field name or a list containing the field names of the fields used to group columns.
     @type  col_fields:    str/list
     @param row_fields:    A single field name or a list containing the field names of the fields used to group rows
     @type  row_fields:    str/list
     @param expressions:   A single expression or a list containing expressions used to do calculations on groupings; when the expression is evaluated, the keyword 'group' denotes the matching records for each cell
     @type  expressions:   str/list
     @return:              A new table containing the crosstabled results.  The last column and last row of the table contain the row and column totals
     @rtype:               Table
  '''
  if isinstance(table, TableArray):
    return run_tablearray('Running pivot...', pivot_table, table, col_fields, row_fields, expressions)
  # table and field type checking is performed in Crosstable.__init__
  try:
    ct = Crosstabler(table, row_fields, col_fields, expressions)
    rowlabels = ct.get_row_labels()
    rowlabels.sort()
    collabels = ct.get_col_labels()
    collabels.sort()
    
    # header (the columns)
    show_progress('Running pivot...', 0.0/len(rowlabels))
    columns = [ make_valid_variable('_'.join(_ensure_tuple(row_fields))) ] # first column is for the row keys
    for collabel in collabels:
      colname = make_valid_variable('_'.join([ unicode(name) for name in _ensure_tuple(collabel)]))
      columns.append(ensure_unique_list_value(columns, colname))
    totalsname = ensure_unique_list_value(columns, 'Totals')
    columns.append(totalsname)
    table = Table(zip(make_unique_colnames(columns), [ unicode for i in columns ]))
  
    # add the data by row
    for k, rowlabel in enumerate(rowlabels):
      show_progress('Running pivot...', float(k)/len(rowlabels))
      # add the row key (first column)
      columns = [ '_'.join([ unicode(name) for name in _ensure_tuple(rowlabel)]) ]
      # add the data cells
      for collabel in collabels:
        if len(ct.expressions) == 1: columns.append(ct.get_cell_calculations(rowlabel, collabel)[0])
        else: columns.append(ct.get_cell_calculations(rowlabel, collabel))
      # add the row subtotal
      if len(ct.expressions) == 1: columns.append(ct.get_row_calculations(rowlabel)[0])
      else: columns.append(ct.get_row_calculations(rowlabel))
      # append the columns to the rows list
      table.append(columns)
      
    # add a row for the column totals
    columns = [ totalsname ]
    for collabel in collabels:
      if len(ct.expressions) == 1: columns.append(ct.get_col_calculations(collabel)[0])
      else: columns.append(ct.get_col_calculations(collabel))
    # add the grand total
    if len(ct.expressions) == 1: columns.append(ct.get_table_calculations()[0])
    else: columns.append(ct.get_table_calculations())
    table.append(columns)
    
    # return the table
    return table
    
  finally:
    clear_progress()

  

def pivot_map(table, col_fields, row_fields, expressions):
  '''Matches all unique combinations of values in col_fields and row_fields with
     expressions run on the records that containing those unique values.
  
     This function is very similar to Grouping.summarize_by_value, only it is formatted
     in a way to make crossstabling possible.

     This is an advanced function that is used internally during the pivot technique.
     It is provided for advanced users who want to access the detail records during
     the crosstabling process.  It is one step in the process beyond pivot_map_detail.

     Most users should use Crosstable.pivot instead as it is more like Excel's pivot function.
     
     Example:
       >>> # create a test table
       >>> data = Table([
       ...  ('Region', unicode),
       ...  ('Product', unicode),
       ...  ('Salesperson', unicode), 
       ...  ('Customer', unicode),
       ...  ('Amount', int),
       ... ],[
       ...   [ 'Rural','Computers',  'Mollie', 'Faiz',  500 ],
       ...   [ 'City',  'Monitors',  'Danny',  'Sheng', 700 ],
       ...   [ 'Rural', 'Mice',      'Mollie', 'Brian', 900 ],
       ...   [ 'City',  'Computers', 'Danny',  'Faiz',  300 ],
       ...   [ 'Rural', 'Monitors',  'Mollie', 'Sheng', 500 ],
       ...   [ 'Rural', 'Monitors',  'Mollie', 'Sheng', 500 ],
       ...   [ 'City',  'Mice',      'Danny',  'Brian', 100 ],
       ...   [ 'Rural', 'Monitors',  'Mollie', 'Faiz',  200 ],
       ...   [ 'City',  'Computers', 'Danny',  'Sheng', 400 ],
       ... ])
       >>> # perform the pivot
       >>> ret = Crosstable.pivot_map(data, 'Salesperson', 'Product', 'sum(group["Amount"])')
       >>> ret
       {
         ('Monitors', 'Mollie'): (1200,), 
         ('Computers', 'Mollie'): (500,), 
         ('Mice', 'Danny'): (100,), 
         ('Computers', 'Danny'): (700,), 
         ('Mice', 'Mollie'): (900,), 
         ('Monitors', 'Danny'): (700,),
       }     
     
     @param table:         The Picalo table that will be crosstabled
     @type  table:         Table
     @param col_fields:    A single field name or a list containing the field names of the fields used to group columns.
     @type  col_fields:    str/list
     @param row_fields:    A single field name or a list containing the field names of the fields used to group rows
     @type  row_fields:    str/list
     @param expressions:   A single expression or a list containing expressions used to do calculations on groupings; when the expression is evaluated, the keyword 'group' denotes the matching records for each cell
     @type  expressions:   str/list
     @return:              A dictionary of each unique key (made up of row and column combinations) mapped to their matching records.
     @rtype:               dict
  '''
  if isinstance(table, TableArray):
    return run_tablearray('Running pivot...', pivot_map, table, col_fields, row_fields, expressions)
  # table and field type checking is performed in Crosstable.__init__
  try:
    # note that this is almost exactly the same code as pivot_map_detail, except it runs get_cell_calculations
    ct = Crosstabler(table, row_fields, col_fields, expressions)
    dict = {}
    rowlabels = ct.get_row_labels()
    for k, rowlabel in enumerate(rowlabels):
      show_progress('Running pivot...', float(k)/len(rowlabels))
      for collabel in ct.get_col_labels():
        dict[(rowlabel, collabel)] = ct.get_cell_calculations(rowlabel, collabel)
    return dict
  finally:
    clear_progress()
    
 
def pivot_map_detail(table, col_fields, row_fields):
  '''Matches all unique combinations of values in col_fields and row_fields with
     Tables that contain only the records with those values.  This function is a
     "mega-select" function that is the basis of the crosstabling technique.
  
     Use this function if you just want to separate a Table in a number of 
     subtables -- one for each unique combination of col_fields and row_fields.
     
     This function is very similar to Grouping.stratify_by_value, only it is formatted
     in a way to make crossstabling possible.

     This is an advanced function that is used internally during the pivot technique.
     It is provided for advanced users who want to access the detail records during
     the crosstabling process.  It is the first step performed in a crosstable.

     Most users should use Crosstable.pivot instead as it is more like Excel's pivot function.
     
     Example:
       >>> # create a test table
       >>> data = Table([
       ...  ('Region', unicode),
       ...  ('Product', unicode),
       ...  ('Salesperson', unicode), 
       ...  ('Customer', unicode),
       ...  ('Amount', int),
       ... ],[
       ...   [ 'Rural','Computers',  'Mollie', 'Faiz',  500 ],
       ...   [ 'City',  'Monitors',  'Danny',  'Sheng', 700 ],
       ...   [ 'Rural', 'Mice',      'Mollie', 'Brian', 900 ],
       ...   [ 'City',  'Computers', 'Danny',  'Faiz',  300 ],
       ...   [ 'Rural', 'Monitors',  'Mollie', 'Sheng', 500 ],
       ...   [ 'Rural', 'Monitors',  'Mollie', 'Sheng', 500 ],
       ...   [ 'City',  'Mice',      'Danny',  'Brian', 100 ],
       ...   [ 'Rural', 'Monitors',  'Mollie', 'Faiz',  200 ],
       ...   [ 'City',  'Computers', 'Danny',  'Sheng', 400 ],
       ... ])
       >>> # perform the pivot
       >>> ret = Crosstable.pivot_map_detail(data, 'Salesperson', 'Product')
       >>> ret
       {
         ('Mollie', 'Computers'): <Table: 1 rows x 5 cols>, 
         ('Danny', 'Computers'):  <Table: 2 rows x 5 cols>, 
         ('Mollie', 'Monitors'):  <Table: 3 rows x 5 cols>, 
         ('Danny', 'Monitors'):   <Table: 1 rows x 5 cols>, 
         ('Danny', 'Mice'):       <Table: 1 rows x 5 cols>, 
         ('Mollie', 'Mice'):      <Table: 1 rows x 5 cols>
       }

     
     @param table:         The Picalo table that will be crosstabled
     @type  table:         Table
     @param col_fields:    A single field name or a list containing the field names of the fields used to group columns.
     @type  col_fields:    str/list
     @param row_fields:    A single field name or a list containing the field names of the fields used to group rows
     @type  row_fields:    str/list
     @return:              A dictionary of each unique key (made up of row and column combinations) mapped to Tables containing their matching records.
     @rtype:               dict
  '''
  if isinstance(table, TableArray):
    return run_tablearray('Running pivot...', pivot_map_detail, table, col_fields, row_fields)
  # table and field type checking is performed in Crosstable.__init__
  try:
    # note that this is almost exactly the same code as pivot_map, except it runs get_cell_records
    ct = Crosstabler(table, row_fields, col_fields, [])
    dict = {}
    rowlabels = ct.get_row_labels()
    for k, rowlabel in enumerate(rowlabels):
      show_progress('Running pivot...', float(k)/len(rowlabels))
      for collabel in ct.get_col_labels():
        dict[(rowlabel, collabel)] = ct.get_cell_records(rowlabel, collabel)
    return dict
  finally:
    clear_progress()
  
  
  
  
############################################################
###   Crosstabler class -- used internally in the functions

class Crosstabler:
  '''Crosstables data sets.  This is similar to Excel's PivotTable feature, although
     somewhat more powerful (although admittedly harder to use).
     
     The Crosstable object is not normally used directly.  The pivot, pivot_source,
     and pivot_map functions should be used instead.
     
     The Crosstable object is live, meaning that it calculates results just in time.
     Any changes to the table are reflected immediately.
  '''
  def __init__(self, table, row_fields, col_fields, expressions):
    '''Sets up the crosstable.  Don't use this object directly.  Call pivot, pivot_source,
       and pivot_map instead.
    
       @param table:        the source Picalo table (list of rows (which are lists themselves))
       @type  table:        Table
       @param row_fields:   a list containing the field names of the fields used to group rows
       @type  row_fields:   list
       @param col_fields:   a list containing the field names of the fields used to group columns
       @type  col_fields:   list
       @param expressions:  a list containing expressions for groupings; for each matching col/row value, each expression is run with the matching records in the "group" variable. 
       @type  expressions:  list
    '''
    # save the information, ensuring we have lists for the col, row, and expressions
    self.table = table
    self.row_fields = _ensure_tuple(row_fields)
    check_valid_table(table, *self.row_fields)
    self.col_fields = _ensure_tuple(col_fields)  
    check_valid_table(table, *self.col_fields)
    self.cell_fields = self.row_fields + self.col_fields
    self.expressions = [ PicaloExpression(expression) for expression in _ensure_tuple(expressions) ]


  def get_col_labels(self):
    '''Returns the unique labels for the column names.  If col_fields was a single value,
       the labels are single values.  If col_fields was a list of values, the labels
       are tuples of values.'''
    return self.table.index(*self.col_fields).keys()
    
    
  def get_row_labels(self):
    '''Returns the unique labels for the column names.  If row_fields was a single value,
       the labels are single values.  If row_fields was a list of values, the labels
       are tuples of values.'''
    return self.table.index(*self.row_fields).keys()
    
    
  def _get_records(self, indexfields, key):
    '''Returns the records that match the given key in the given index'''
    index = self.table.index(*indexfields)  # this is an efficient method, see the Table class.  It also automatically updates when the table changes.
    if len(indexfields) == 1:  # Table.index doesn't use tuples for 1-length tuple keys
      if index.has_key(key[0]):
        return Simple.select_records(self.table, index[key[0]])
      return Simple.select_records(self.table, [])
    else:
      if index.has_key(key):
        return Simple.select_records(self.table, index[key])
      return Simple.select_records(self.table, [])
    
    
  def _get_calculations(self, indexfields=None, key=None):
    '''Returns the calculations for records that match the given key in the given index'''
    if indexfields == None or key == None:  # if indexfields or key is None, then we're doing the entire table
      group = self.table
    else:
      group = self._get_records(indexfields, key)
    ret = []
    for exp in self.expressions:
      ret.append(exp.evaluate([{'group':group}]))
    return tuple(ret)
    
    
  def get_cell_records(self, rowlabel, collabel):
    '''Returns the records that match the given rowlabel and collabel unique values'''
    return self._get_records(self.cell_fields, _ensure_tuple(rowlabel) + _ensure_tuple(collabel))
    
    
  def get_cell_calculations(self, rowlabel, collabel):
    '''Returns the cell calculations for the cell with the given row label unique value
       and column label unique value'''
    return self._get_calculations(self.cell_fields, _ensure_tuple(rowlabel) + _ensure_tuple(collabel))
    
    
  def get_row_records(self, rowlabel):
    '''Returns the records that match the given rowlabel unique value'''
    return self._get_records(self.row_fields, _ensure_tuple(rowlabel))

    
  def get_row_calculations(self, rowlabel):
    '''Returns the cell calculations for the row with the given row label'''
    return self._get_calculations(self.row_fields, _ensure_tuple(rowlabel))
    
    
  def get_col_records(self, collabel):
    '''Returns the records that match the given collabel unique value'''
    return self._get_records(self.col_fields, _ensure_tuple(collabel))

    
  def get_col_calculations(self, collabel):
    '''Returns the cell calculations for the col with the given row label'''
    return self._get_calculations(self.col_fields, _ensure_tuple(collabel))
    
    
  def get_table_calculations(self):
    '''Returns the grand total calculations for the entire table'''
    return self._get_calculations()
    
    
    
#########################################
###   Helper functions

def _ensure_tuple(var):
  '''Ensures the given variable is a tuple'''
  if isinstance(var, types.TupleType):
    return var
  elif isinstance(var, types.ListType):
    return tuple(var)
  return tuple((var,))  # without the comma, strings are split into characters
  
  
      



#!/usr/bin/python

####################################################################################
#                                                                                  #
# Copyright (c) 2003 Dr. Conan C. Albrecht <conan_albrechtATbyuDOTedu>             #
#                                                                                  #
# This file is part of Picalo.                                                     #
#                                                                                  #
# Picalo is free software; you can redistribute it and/or modify                   #
# it under the terms of the GNU General Public License as published by             #
# the Free Software Foundation; either version 2 of the License, or                # 
# (at your option) any later version.                                              #
#                                                                                  #
# Picalo is distributed in the hope that it will be useful,                        #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                   #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                    #
# GNU General Public License for more details.                                     #
#                                                                                  #
# You should have received a copy of the GNU General Public License                #
# along with Foobar; if not, write to the Free Software                            #
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA        #
#                                                                                  #
####################################################################################
#                                                                                  
# The modules in the base/ directory are globally imported into the Python namespace
# when you run "from picalo import *".  They are separated here into different files
# for convenience (rather than putting it all in __init__.py).  See picalo/__init__.py
# for how they are globally imported into Picalo.
#                                                                                  
####################################################################################

# this documentation is placed on the top page of the epydoc API documentation
__doc__ = '''
Picalo is a Python library to help anyone who works with data files, 
especially those who work with data in relational/spreadsheet format.  
It is primarily created for investigators and auditors search through 
data sets for anomalies, trends, ond other information, but it is 
generally useful for any type of data or text files.

Picalo is different from NumPy/Numarray in that it is meant for
heterogeneous data rather than homogenous data.  In NumPy, you
have an array (table) of the same type--all ints, for example.
In Picalo, you have a table made up of different column types,
very similar to a database.

One of Picalo's primary purposes is making relational
databases easier to work with.  Once you have a Picalo table, 
you can add, move, or delete columns; work with records (horizontal
slices of the data); select and group records in various ways;
and run analyses on tables.  Picalo includes adapters for popular
databases, and it provides a Query object that make queries seem
just like regular Tables (except they are live from the database).

If you work with relational databases, delimited (CSV/TSV) files, 
EBCDIC files, MS Excel files, log files, text files, or other 
heterogeneous datasets, Picalo might make your life easier.

Picalo is programmed to be as Pythonic as possible.  It's core objects--
tables, columns, records--they act like lists.  A column is a list of cells.
A record is a list of cells.  A table is a list of records.  Tables can be 
sorted via the sort function, just like the Sorting HowTo shows.  The return
values of almost all functions are new tables, so functions can be chained
together like pipes in Unix.

Picalo includes an optional Project object that stores tables in
Zope Object DB files.  When Projects are used, Picalo automatically
swaps records in and out of memory as needed to ensure efficient use of 
resources.  Projects allow Picalo to work with essentially an unlimited
amount of data.

The project was started in 2003 by Conan C. Albrecht, a professor
in Information Systems at Brigham Young University.  Conan remains
the primary developer of Picalo.

Picalo is released in two formats.  First, as a pure-Python library that is used 
by simply "import picalo" or "from picalo import *" in any Python script.  
Python programmers will be primarily interested in the library
version.  This format is installed in the typical Python fashion, either
as an .egg via setuptools, or via "python setup.py install" from
the source.

Second, Picalo is released as a standalone, wx-Python-based GUI environment that allow
end users to access the Picalo libraries. This version is packaged as a Windows setup.exe file, Mac
application bundle, and Linux rpm and deb files.  The user
may not realize Python is even being used when running the
full application environment.
</ol>

Enjoy!  Please report any bugs to me.  I also welcome additions to the toolkit.
  
Dr. Conan C. Albrecht
conan@warp.byu.edu
http://www.picalo.org/
'''



########################################################
###   Imports

import sys, types, re, inspect, os, os.path, threading, decimal



######################################################
###   Value formatting and parsing routines

format_cache = {}
format_parser = re.compile('^([^#0]*?)([#0]*?),{0,1}([0#]*)\.{0,1}(0*)(%{0,1})(e\+(0+)){0,1}$', re.IGNORECASE)

class NumberFormat:
  def __init__(self, format):
    '''Initializes the NumberFormat from the given format'''
    self.prefix = None
    self.decimalplaces = None
    self.commapos = None
    self.percent = False
    self.sciplaces = None
    result = format_parser.search(format.strip())
    assert result != None, 'Invalid number format: %s' % format
    if result.group(1):
      self.prefix = result.group(1)
    if result.group(2) and result.group(3):
      self.commapos = len(result.group(3))
    if result.group(3) and result.group(3)[-1] == '0':
      self.decimalplaces = 0
    if result.group(4):
      self.decimalplaces = len(result.group(4))
    if result.group(5):
      self.percent = True
    if result.group(6):
      self.sciplaces = len(result.group(7))

    
  def format_value(self, value, typ):
    '''Returns the value as a formatted string'''
    # scientific notation
    if self.sciplaces != None:
      val = float(value)
      exp = 0
      if abs(val) >= 1:
        while abs(val) >= 10:
          val = val / 10
          exp += 1
        return ('%0.' + unicode(self.sciplaces) + 'fE+%i') % (val, exp)
      else:  # less than 1
        while abs(val) < 1:
          val = val * 10
          exp += 1
        return ('%0.' + unicode(self.sciplaces) + 'fE-%i') % (val, exp)
    
    # if we're adding a percent, times the number by 100
    value2 = decimal.Decimal(unicode(value))
    if self.percent:
      value2 = value2 * 100
      
    # round the number to the specified number of decimal places
    # the decimal.quantize method isn't working for some reason, so I'll temporarily use round for now
    if self.decimalplaces != None:
      value2 = decimal.Decimal(unicode(round(value2, self.decimalplaces)))

    # get the parts of the number
    parts = value2.as_tuple()
    if parts[2] == 0:
      intpart = [ unicode(s) for s in parts[1] ]
      decpart = []
    else:
      intpart = [ unicode(s) for s in parts[1][:parts[2]] ]
      decpart = [ unicode(s) for s in parts[1][parts[2]:] ]
    
    # do the integer part
    if self.commapos:
      i = len(intpart) - 1
      x = 0
      while i > 0:
        x += 1
        if x  % self.commapos == 0:
          intpart.insert(i, ',')
        i -= 1
      ret = ''.join(intpart)
    else:
      ret = unicode(int(value))
    
    # do the decimal part
    if self.decimalplaces == None and not typ in (int, long):
      ret += '.' + ''.join([ unicode(s) for s in decpart ])
    elif self.decimalplaces > 0:
      while len(decpart) < self.decimalplaces:
        decpart.append('0')
      ret += '.' + ''.join([ unicode(s) for s in decpart ])
      
    # add the percent symbol if we've been asked to
    if self.percent:
      ret += '%'
    
    # add the prefix
    if self.prefix:
      ret = self.prefix + ret
      
    # return the string
    return ret
    

def format_value_from_type(value, typ, format=None):
  '''Utility method that does the actual formatting.  See the Column.set_format method
     for a description of the format used here.'''
  # short circuit if there's an error here
  if isinstance(value, error):
    return unicode(value)
    
  # if the None value
  if value == None:
    return '<N>'
     
  # if a date format
  if typ in (Date, DateTime):
    if format != None:
      return value.strftime(format)
    elif typ == Date:
      return value.strftime("%Y-%m-%d")
    elif typ == DateTime:
      return value.strftime("%Y-%m-%d %H:%M:%S")
  
  # if a number format
  if typ in (int, long, float, number, currency):
    if format:
      # get this format object from the cache
      if format_cache.has_key(format):
        fmt = format_cache[format]
      else:
        fmt = NumberFormat(format)
        format_cache[format] = fmt
      # return the formated value
      return fmt.format_value(value, typ)
    else:   # just use the type's native format
      return unicode(value)
  
  # if any other format, just convert to a string
  return unicode(value)
  
  
# regular expressions used in parsing numbers
number_parser_main = re.compile('[^-0-9.]', re.DOTALL)
number_parser_dot_remover = re.compile('^(.*?\..*?)\.', re.DOTALL)
number_parser_percent = re.compile('^(.*?)%', re.DOTALL)
number_parser_sci_note = re.compile('^(.*?)E([-+])([0-9]+)', re.IGNORECASE | re.DOTALL)

def parse_value_to_type(value, typ, format=None):
  '''Utility method that does the actual parsing.  See the Column.set_format method
     for a description of the format used here.'''
  # short circuit if we're already the right type
  # short circuit if we have the None type (don't ever coerce that one)
  # short circuit if we don't have a column type (which should really never happen)
  if value == None or isinstance(value, typ) or typ == None:
    return value
  
  # if a date format
  if typ == Date:
    if format != None:
      return Date(value, format)
    return Date(value)
  elif typ == DateTime:
    if format != None:
      return DateTime(value, format)
    return DateTime(value)
  
  # if a number format
  if typ in (int, long, float, number, currency):
    num = unicode(value)
    # check for scientific notation, and separate it out if necessary
    result_scinote = number_parser_sci_note.search(num)
    if result_scinote != None:
      num = result_scinote.group(1)

    # check for a percent sign
    result_perc = number_parser_percent.search(num)
    if result_perc:
      num = result_perc.group(1)

    # this removes everything except numbers and dots
    num = number_parser_main.sub('', num)

    # check for multiple dots, and ignore anything after the second dot
    result = number_parser_dot_remover.search(num)
    if result:
      num = result.group(1)

    # if we have parsed everything out of the string, assign a 0 to the value
    if num == '':
      num = 0
  
    # we can now convert to an actual number
    conv = typ(number(num))  # have to go through number because int can't parse a string containing a period
    if result_scinote != None and result_scinote.group(2) == '-':
      conv = typ(conv / (10 ** int(result_scinote.group(3))))
    if result_scinote != None and result_scinote.group(2) == '+':
      conv = typ(conv * (10 ** int(result_scinote.group(3))))
    if result_perc != None:
      conv = typ(conv / 100.0)
    return conv
  
  # if any other format, try to coerce using the type itself (like int("5"))
  return typ(value)
    


########################################################################
###  Imports must come after the above because the above are needed
###  in some of these imports    
from Calendar import RE_DATE_FORMATS, RE_DATETIME_FORMATS, DateTime, Date
from Number import number
from Currency import currency
from Boolean import boolean
from picalo.lib import stats
from picalo.lib import progressbar
from Error import error
    


#########################################################
###  types used below in Database.post_table
###

VARIABLE_TYPES_RE = []
VARIABLE_TYPES_RE.append(  [ re.compile("^-{0,1}\d+$")            ,     int      ])
VARIABLE_TYPES_RE.append(  [ re.compile("^-{0,1}\d+$")            ,     long     ])
VARIABLE_TYPES_RE.append(  [ re.compile("^-{0,1}\d+(\.\d+){0,1}$"),     float    ])
VARIABLE_TYPES_RE.append(  [ re.compile('^TRUE|FALSE$', re.IGNORECASE), boolean  ])
VARIABLE_TYPES_RE.append(  [ re.compile('^T|F$',        re.IGNORECASE), boolean  ])
VARIABLE_TYPES_RE.append(  [ re.compile('^YES|NO$', re.IGNORECASE),     boolean  ])
for item in RE_DATE_FORMATS:
  VARIABLE_TYPES_RE.append([ item.regex, Date ])
for item in RE_DATETIME_FORMATS:
  VARIABLE_TYPES_RE.append([ item.regex, DateTime ])


# conversions from python types to database types
# maps to the database create string and whether to use quotes around values
TYPE_TO_DB = {
  DateTime: 'TIMESTAMP',
  Date:     'DATE',
  int:      'INTEGER',
  long:     'BIGINT',
  float:    'FLOAT',
  boolean:  'BOOLEAN',
  # str and unicode types (VARCHAR) should not be listed here - see Database.post_table
}
  



# regular expression to check for a valid python variable      
VARIABLE_RE = re.compile('^[A-Za-z_][A-Za-z0-9_]*$')
FIRST_LETTER_RE = re.compile('[A-Za-z_]')
LETTER_RE = re.compile('[A-Za-z0-9_]')
RESERVED_COLUMN_NAMES = {  # these would conflict in expressions
  'record': None,
  'recordindex': None,
  'startrecord': None,
  'record1': None,
  'record1index': None,
  'record2index': None,
  'record2': None, 
  'value': None,
  'group': None,
}
# more RESERVED_COLUMN_NAMES are lower in this module
  



###########################################################
###   A few simple functions to help users do common tasks.
###   These override some Python functions like sum, but
###   they take the exact same arguments and call the
###   built-in functions.  Putting them here allows
###   us to create custom documentation.
_builtinsum = sum
def sum(sequence, start=0):
  '''Returns the sum of the given sequence of numbers
     plus the value of start.  When the sequence is empty,
     returns start.
     
     @param sequence:   A sequence of numbers
     @type  sequence:   list
     @param start:      The starting value, usually 0
     @type  start:      int
     @return:           The sum of the sequence, plus the start value
     @rtype:            int
  '''
  return _builtinsum(sequence, start)
  
  
def mean(sequence, default=0):
  '''Returns the average of the given sequence,
     or the default if the sequence is empty.
     More advanced statistical routines can be found in the picalo.lib.stats
     module.
  
     @param sequence:    A sequence of numbers
     @type  sequence:    list
     @param default:     The default value to return when the list is empty
     @type  default:     int
     @return:            The average of the numbers
     @rtype:             float
  '''
  if len(sequence) == 0:
    return default
  elif len(sequence) == 1:
    return sequence[0]
  else:
    return stats.mean(sequence)
    

def count(sequence):
  '''Returns the number of items in the sequence.
     The built in function "len" also gives this value.
     
     @param sequence:    A sequence of items of any type.  If not a sequence, returns 1.
     @type  sequence:    list
     @return:              The number of items in the sequence
     @rtype:               float
  '''
  try:
    return len(sequence)
  except:
    return 1
  

_builtinmax = max
def max(*args, **kargs):
  '''With a single sequence argument, return its largest item.
     With two or more arguments, return the largest argument.
     
     @param args:   A sequence of items.
     @return:       The largest item in the sequence
     @rtype:        object
  '''
  return _builtinmax(*args, **kargs)
  
  
_builtinmin = min
def min(*args, **kargs):
  '''With a single sequence argument, return its smallest item.
     With two or more arguments, return the smallest argument.
     
     @param args:       A sequence of items.
     @return:           The smallest item in the sequence
     @rtype:            object
  '''
  return _builtinmin(*args, **kargs)

  
def stdev(sequence, default=0):
  '''Returns the standard deviation of the given sequence,
     or the default if the sequence contains zero or one items.
     More advanced statistical routines can be found in the picalo.lib.stats
     module.
     
     @param sequence:   A sequence of items.
     @type  sequence:   list
     @param default:    The default if a standard deviation cannot be calculated.
     @type  default:    float
     @return:           The standard deviation of the sequence, or the default if len(sequence) < 2.
     @rtype:            float
  '''
  if len(sequence) < 2:
    return default
  return stats.stdev(sequence)
  
  
def variance(sequence, default=0):
  '''Returns the variance of the given sequence,
     or the default if the sequence contains zero or one items.
     More advanced statistical routines can be found in the picalo.lib.stats
     module.
     
     @param sequence:   A sequence of items.
     @type  sequence:   list
     @param default:    The default if a variance cannot be calculated.
     @type  default:    float
     @return:           The variance of the sequence, or the default if len(sequence) < 2.
     @rtype:            float
  '''
  if len(sequence) < 2:
    return default
  return stats.var(sequence)
  
  
#################################################################################
###   Whether Picalo is being run in GUI mode.
###   This has to be in a module that doesn't import (or indirectly import wx)
###   so Picalo can be used from command line without loading wx all the time

# a reference to the main frame of the GUI program.  PicaloApp.py sets this.
# if None, we're running in console mode.
mainframe = None
guiUpdateProgressDialog = None
useProgress = False  # this is set to True at the end of picalo/__init__.py
lastcaller = None
lastprogress = 0.0
progress_lock = threading.RLock()

def _updateProgress(msg='', progress=1, title='Progress', force=False):
  '''Initializes the progress bar if it's being used.  Progress should be from 0 to 1.
     This function SHOULD NOT be called directly.  Instead, call the show_progress()
     method global to Picalo (imported when you "from picalo import *").
     
     See the documentation for show_progress for information on the parameters.
  '''
  global lastcaller, lastprogress
  progress_lock.acquire()
  try:
    # if we're not using the progress bars, short circuit now
    if not useProgress:
      return
  
    # short circuit if the last frame who called this isn't the current one
    # I do this because so many functions try to show a progress dialog
    # If one opens the progress dialogs, then another tries to, it would blast the
    # original one away.  So the first to open the dialog is the winner, others get
    # short circuited right here
    caller = id(inspect.currentframe().f_back.f_back.f_code)
    if caller != None and lastcaller != None and caller != lastcaller and force == False:
      # still call the progress bar to give the dialog a chance to catch the cancel button
      if mainframe == None:  # text progress bar
        _updateProgressDialog(progress=lastprogress)
      else:  # GUI progress dialog
        guiUpdateProgressDialog(progress=lastprogress, parent=mainframe)
      return
    lastcaller = caller  
    
    # remove the lastcaller if we're done
    if not msg or progress >= 1.0 or progress < 0.0:
      lastcaller = None
  
    # switch to either text or GUI progress bar
    lastprogress = progress
    if mainframe == None:  # text progress bar
      _updateProgressDialog(msg=msg, progress=progress)
    else:  # GUI progress dialog
      guiUpdateProgressDialog(msg=msg, progress=progress, title=title, parent=mainframe)

  finally:
    progress_lock.release()
    

def use_progress_indicators(show):
  '''Sets whether Picalo shows progress dialogs in text or GUI mode.
     Send False into this method to make Picalo quiet.  Send True
     to see progress bars for operations.
  '''
  global useProgress
  useProgress = show
  

def show_progress(msg='', progress=1.0, title='Progress', force=False):
  '''Updates the progress bar with a message and
     a percentage progress between 0 and 1. To remove the progress
     bar, call clear_progress().
     
     This function is important because it gives feedback to the user.
     In addition, and perhaps more importantly, it gives the user a 
     cancel button (in GUI mode) that allows the user to cancel
     your script.  Be sure to call show_progress throughout your
     script.
     
     Sometimes multiple functions try to show or clear a progress bar.
     For example, a top-level script might show a master progress bar
     and then call load().  The load() function tries to show another
     progress bar, which Picalo normally circumvents or the load() function
     would take over the top-level script's progress bar.  In other words,
     the first script to show a progress bar is the only one that can update
     and/or clear the dialog.  By setting force to True, you can override this
     default behavior.  This should not normally be used as it takes control 
     when the top-level script should keep control.

     @param msg:        The message to show the user.
     @type  msg:        str
     @param progress:   A value between 0 and 1 indicating the percentage finished.
     @type  progress:   float
     @param title:      The title of the progress bar.  Defaults to 'Progress'.
     @type  title:      str
     @param force:      Whether to force control of the dialog to the calling code.  
     @type  force:      boolean
  '''
  _updateProgress(msg=msg, progress=progress, title=title, force=force)  # just a convenience function


def clear_progress(force=False):
  '''Clears the progress bar from the screen.  
  
     Sometimes multiple functions try to show or clear a progress bar.
     For example, a top-level script might show a master progress bar
     and then call load().  The load() function tries to show another
     progress bar, which Picalo normally circumvents or the load() function
     would take over the top-level script's progress bar.  In other words,
     the first script to show a progress bar is the only one that can update
     and/or clear the dialog.  By setting force to True, you can override this
     default behavior.  This should not normally be used as it takes control 
     when the top-level script should keep control.

     @param force:      Whether to force control of the dialog to the calling code.  
     @type  force:      boolean
  '''
  _updateProgress(force=force)  # just a convenience function



###################################
###   Text-based progress bar


class MessageProgressBarWidget(progressbar.ProgressBarWidget):
  '''Extension to display a message'''
  def __init__(self):
    self.msg = ''
  def update(self, pbar):
    return self.msg
message_widget = MessageProgressBarWidget()
    
widgets = (
  message_widget,
  '  ',
  progressbar.Percentage(), 
  ' ', 
  progressbar.Bar(marker="#", left="|", right="|"), 
  ' ', 
  progressbar.ETA() 
)

progressDialog = None

def _updateProgressDialog(msg='', progress=1):
  '''Initializes the progress bar.  This should not be called directly.
     Call show_progress() instead, which gets imported when you run
     "from picalo import *".
     '''
  global progressDialog
  # if progress is done, hide the dialog
  if progress >= 1.0 or progress < 0.0:
    if progressDialog:
      progressDialog.finish()
    progressDialog = None
    return

  # create the progress dialog if necessary
  if not progressDialog:
    progressDialog = progressbar.ProgressBar(maxval=100, widgets=widgets, fd=sys.stdout)
    progressDialog.start()
    
  # update the message if necessary
  if msg:
    message_widget.msg = msg

  # update the progress
  progressDialog.update(int(progress*100.0))
  sys.stdout.flush()




  
##################################
###   Assertion code for errors

def check_valid_table(table, *columns):
  '''Checks to ensure the table is a valid table object, and that the columns
     are valid columns in the table.  Throws an AssertionError if anything is
     wrong.'''
  assert isinstance(table, Table), 'Please specify a valid table for this function.'
  for column in columns:
    if isinstance(column, (list, tuple)):
      check_valid_table(table, *column)
    else:
      table.deref_column(column) # checks that the column index/name is valie


def is_valid_variable(varname):
  '''Returns whether the given varname is a valid python variablename'''
  return VARIABLE_RE.match(varname) != None


def make_valid_variable(varname, repl='_'):
  '''Makes the given variable a valid variable name'''
  # ensure a string
  if not isinstance(varname, types.StringTypes):
    varname = unicode(varname)
  # if empty
  if len(varname) == 0:
    varname = repl
  # check the first variable
  if not FIRST_LETTER_RE.match(varname[0]):  
    if LETTER_RE.match(varname[0]): # if the letter works as a second character, just add a _ on to the first, this often prevents a larger change
      varname = repl + varname
    else:  
      varname = repl + varname[1:]
  # check the rest of the characters and replace each letter that doesn't match with a _
  for i in range(1, len(varname)):
    if not LETTER_RE.match(varname[i]):
      varname = varname[:i] + repl + varname[i+1:]
  return varname
  

def ensure_valid_variables(lst, repl='_'):
  '''Ensures the strings in the list are valid Picalo/Python variables.
     This is used to create column names during loading picalo tables.
     
     This method modifies the lst directly.  It also returns the lst for
     convenience reasons.
  '''
  for i in range(len(lst)):
    lst[i] = make_valid_variable(lst[i], repl)
  return lst
  
  
  
#############################################################
###  Helps create new column names that are unique for 
###  a table.  This is used in various picalo functions
###  that need to calculate unique names.

def make_unique_colnames(columns):
  '''Takes a list of names and adds the appropriate values to
     each value to ensure each is unique'''
  unique = []
  for i in range(len(columns)):
    unique.append(ensure_unique_list_value(unique, columns[i]))
  return unique
  

def ensure_unique_colname(table, name):
  '''Ensures the name is unique for the columns in the table.  It
     adds _1, _2, _3, and so forth if needed to the name.
  '''
  colnames = [ col.name for col in table.get_columns() ]
  return ensure_unique_list_value(colnames, name)
  
  
def ensure_unique_list_value(lst, name):
  '''Ensures that the name is unique for the names
     already in the list.  It adds _1, _2, _3, and so forth
     if needed to the name.  This function is split out
     because functions like Crosstable need to create the
     column name list before creating the table.
  '''
  ret = name
  i = 1
  while ret in lst:
    ret = name + '_' + str(i)
    i += 1
  return ret
  
  
  
##############################################
###   Converts any value to unicode using utf-8

def make_unicode(value):
  '''Makes a value (of any type) into unicode using utf_8'''
  if isinstance(value, unicode):
    return value
  elif isinstance(value, str):
    return unicode(value, 'utf_8')
  else:
    return unicode(repr(value), 'utf_8')



###########################################
###   Runs functions for table lists

def run_tablearray(msg, func, tablearray, *args, **kargs):
  '''Runs the given function on every table in the TableArray.  If the 
     function returns a Table, the results are collected into another
     TableList and returned.
     
     It assumes that the first parameter in the function is the table.
  '''
  # we assume the types have already been checked since this method is called
  # only by Picalo code (not by end users)
  # first put the tablelist into the arguments
  try:
    results = TableArray()
    return_results = False
    for i, table in enumerate(tablearray):
      show_progress(msg, float(i) / float(len(tablearray)))
      ret = func(table, *args, **kargs)
      if isinstance(ret, (Table, TableArray, TableList)):
        return_results = True
        results.append(ret)
    if return_results:
      return results
  finally:
    clear_progress()
  
  
  
  
  


###########################################################
###   General helper methods

def create_directory(directory):
  '''Creates the given directory, including all required
     directories.  This works equally well on Windows 
     and Unix (the os.mkdirs doesn't seem to like c: in paths).
  '''
  # figure out how much of the path needs to be created
  parts = []
  lastdir = None # to detect an infinite loop -- where os.path.split(directory) == directory
  while not os.path.exists(directory) and lastdir != directory:
    lastdir = directory
    directory, tail = os.path.split(directory)
    if tail:
      parts.append(tail)
  parts.reverse()
  
  # go through each one and make the directory
  for part in parts:
    directory = os.path.join(directory, part)
    os.mkdir(directory)
  
  
  
#####################################################
###   An abstract decorator for any class

class AbstractDecorator:
  '''An abstract decorator that directs all attribute/method access first
     here, then to the decorated object, then back here (for the first time something is set).
     
     Subclasses *must* call AbstractDecorator.__init__ before setting any self. variables or
     other attributes (so the _decorated_object can be set up).
  '''
  def __init__(self, obj):
    self._decorated_object = obj
    
  def __getattr__(self, name):
    try:
      return self.__dict__[name]
    except KeyError:
      return getattr(self._decorated_object, name)
    
  def __setattr__(self, name, value):
    if name == '_decorated_object' or name in self.__dict__.keys() or not hasattr(self._decorated_object, name):
      self.__dict__[name] = value
    else:
      setattr(self._decorated_object, name, value)
    
  def __delattr__(self, name):
    if name == '_decorated_object' or name in self.__dict__.keys() or not hasattr(self._decorated_object, name):
      del(self.__dict__[name])
    else:
      delattr(self._decorated_object, name)
      
  # is there a better way to do this?
  def __str__(self): return str(self._decorated_object)
  def __repr__(self): return repr(self._decorated_object)
  def __eq__(self, other): return self._decorated_object == other
  def __ne__(self, other): return self._decorated_object != other
  def __lt__(self, other): return self._decorated_object < other
  def __le__(self, other): return self._decorated_object <= other
  def __gt__(self, other): return self._decorated_object > other
  def __ge__(self, other): return self._decorated_object >= other
  def __cmp__(self, other): return cmp(self._decorated_object, other)
  def __hash__(self): return hash(self._decorated_object)
  def __nonzero__(self): return self._decorated_object != None


  
# now that we have some constants and basic classes defined, import more modules (that need the constants above)
from Table import Table
from TableList import TableList
from TableArray import TableArray
from Record import Record
from Column import Column


    
# add all record methods to the reserved column names list
# keep the methods and class variables in Record low because we allow record.column access
# and each method added further limits the available column names the users can use
table = Table(['id'], [[0]])  # get a table so we can do a dir() on a Record object
RESERVED_COLUMN_NAMES.update([ (name, None) for name in dir(table[0]) ])

  
  
##############################################################################################################
###   Alter the stats functions to accept
###   Columns and Records in addition to
###   Lists and Tuples.  That way Picalo objects
###   can be inputs into the stats functions

# this is a bit invasive and expects exact structures within stats, 
# so check that we are using the expected version of stats.
assert stats.__version__ == 0.6, 'The picalo.lib.stats module has been upgraded.  The Picalo developers need to modify the picalo.lib.__init__ file to match it.'

# add Column and Record to every function that accepts a Python list
for name in dir(stats):
  func = getattr(stats, name)
  if isinstance(func, stats.Dispatch):
    if types.ListType in func._dispatch:
      func._dispatch[Column] = func._dispatch[types.ListType]
      func._dispatch[Record] = func._dispatch[types.ListType]
      


##################################################################################
###   Import the other Picalo files and objects into the "picalo." namespace

# the order here is important!
# the from xxx import * is important for "from picalo import *" syntax
import Boolean as BooleanModule
from Boolean import *
import Number as NumberModule
from Number import *
import Currency as CurrencyModule
from Currency import *
import Error as ErrorModule
from Error import *
import Calendar as CalendarModule
from Calendar import *
import Filer as FilerModule
from Filer import *
import Table as TableModule
from Table import *
import TableArray as TableArrayModule
from TableArray import *
import TableList as TableListModule
from TableList import *
import Project as ProjectModule
from Project import *
global_variables =         \
  BooleanModule.__all__    +\
  NumberModule.__all__     +\
  CurrencyModule.__all__   +\
  ErrorModule.__all__      +\
  CalendarModule.__all__   +\
  FilerModule.__all__      +\
  TableModule.__all__      +\
  TableArrayModule.__all__ +\
  TableListModule.__all__  +\
  ProjectModule.__all__      

# this part is required for the "import picalo" use
import Benfords
import Crosstable
import Financial
import Grouping
import Trending
import Simple  
  
# modules that load with picalo.  these modules must be used with
# their names, as in "Simple.describe" rather than simply "describe".
global_modules = [
  'Benfords',
  'Boolean',
  'Crosstable',
  'Currency',
  'Column',
  'Calendar',
  'Database',
  'Error',
  'Expression',
  'Financial',
  'Grouping',
  'Number',
  'Filer',
  'Table',
  'Project',
  'Record',
  'Simple',
  'TableArray',
  'TableList',
  'Trending',
]

# the gui imports these into the shell automatically for the user
# yes, I arbitrarily decided which ones most users would want always
# included based on my experience with quite a few users using Picalo
python_modules = [
  'string',
  'sys',
  're',
  'random',
  'os',
  'os.path',
  'urllib',
  'xml.etree.ElementTree',
  
]


# global functions (defined in this file)
global_functions = [
  'sum',
  'count',
  'mean',
  'max',
  'min',
  'stdev',
  'variance',
  'use_progress_indicators',
  'show_progress',
  'clear_progress',
  'check_valid_table',
  'format_value_from_type',
  'parse_value_to_type',
]


# export the variables and modules in the __all__ variable
__all__ = global_variables + global_modules + global_functions



###############################################################################################
###   Final setup  

# now that we're done setting up, show progress indicators for user code 
use_progress_indicators(True)





